# -*- coding: utf-8 -*-
"""
"""
import os
import sys

# Log file location. Log rotations/archives will be put in LOG_DIR.
LOG_FILE = 'pynuget.log'
LOG_DIR = "/var/log/pynuget"
if sys.platform == 'win32':
    LOG_DIR = 'C:/Logs/pynuget'
LOG_PATH = os.path.normpath(os.path.join(LOG_DIR, LOG_FILE))

# The database backend to use.
# Must be ond of ("sqlite", "mysql", "postgresql")
DB_BACKEND = "sqlite"

# Schema name or path to SQLite file.
DB_NAME = "nuget_packages.sqlite"

# Server path: root path for things like the SQLite file and the packge files.
# Must be asolute.
### XXX DO NOT CHANGE THIS XXX ###
# I haven't spent the time needed in order to have a custom install dir...
SERVER_PATH = "/var/www/pynuget"
### XXXXXXXXXXXXXXXXXXXXXXXXXX ###

# Where nuget packages will be saved to and sourced from.
# Can be absolute or relative. Defaults to $SERVER_PATH\$PACAKGE_DIR
PACKAGE_DIR = "nuget_packages"

# The name of the Apache configuration file
APACHE_CONFIG = "pynuget.conf"

# A list (well, set) of valid API Keys
API_KEYS = {
    "ChangeThisKey",
}

# These are only used for non-SQLite backends like MySQL or PostgreSQL.
DB_USER = None
DB_PASSWORD = None



# -*- coding: utf-8 -*-
"""
"""
import os
import re
import shutil
import subprocess
import sys
from datetime import datetime as dt
from pathlib import Path

import requests
from sqlalchemy import create_engine

from pynuget import db
from pynuget import _logging


logger = _logging.setup_logging(True, False, "./pynuget-cli.log")


def init(server_path, package_dir, db_name, db_backend, apache_config,
         replace_wsgi=False, replace_apache=False):
    """
    Initialize the PyNuGet server.

    Parameters
    ----------
    server_path : str
        The directory that you want the server to live in. Must be an
        absolue path. Defaults to /var/www/pynuget
    package_dir : str
        The directory that the packages will be saved in and served from. If
        this value is not an absolute path, `server_path` will be used as
        the root. Defaults to 'nuget_packages' in `server_path`
    db_name : str
        The name of the database to use. If db_backend == 'sqlite', then
        this is the relative or absolute path to the SQLite file to use. If
        db_backend != 'sqlite', then this is the name of the Schema to create.
        Defaults to 'packages.db' in `server_path`.
    db_backend : str
        One of ('sqlite', 'mysql', 'postgresql'). Defaults to 'sqlite'.
    apache_config : str
        The name of the apache configuration file. Defaults to 'pynuget.conf'.

    Details
    -------
    +  Create /var/www/pynuget if it doesn't exist.
    +  Create /var/www/pynuget/package_files if it doesn't exist
    +  Create /var/www/pynuget/wsgi.py if it doesn't exist
    +  Copy Apache config if it doesn't exist
    +  Enable Apache site.
    +  Create the DB file or schema if it doesn't exist.
    """
    args = dict(locals())
    _check_permissions()
    _create_directories(server_path, package_dir, "/var/log/pynuget")
    _create_db(db_backend, db_name, server_path)

    wsgi_file = _copy_wsgi(server_path, replace_wsgi)
    _update_wsgi(wsgi_file)

    conf = _copy_apache_config(apache_config, replace_apache)
    _enable_apache_conf(conf.resolve())

    default_config_file = Path(__file__).parent / "default_config.py"
    _save_config(default_config_file, **args)

    _set_permissions([server_path, "/var/log/pynuget"])

    _reload_apache()

    return True


def clear(server_path, force=False):
    """
    +  Truncate/drop all tables and recreate them
    +  Delete all packages in package_files
    """
    # Confirm:
    ans = input("Are you sure you want to delete all packages? [yN]")
    if ans.lower() in ('y', 'yes'):
        pass
    elif ans.lower() in('n', 'no'):
        logger.debug("User aborted.")
        sys.exit(0)
    else:
        logger.debug("Unknown response '%s'. Aborting." % ans)
        sys.exit(1)

    # Read the config file to find our other locations
    # TODO: I hate this...
    sys.path.append(server_path)
    import config

    # Delete all the packages
    pkg_path = Path(config.SERVER_PATH) / Path(config.PACKAGE_DIR)
    try:
        shutil.rmtree(str(pkg_path))
    except FileNotFoundError:
        logger.warn("Path '%s' does not exist." % str(pkg_path))

    # Delete/drop the database
    if config.DB_BACKEND == 'sqlite':
        sqlite_path = Path(config.SERVER_PATH) / Path(config.DB_NAME)
        try:
            sqlite_path.unlink()
        except FileNotFoundError:
            logger.warn("Path '%s' does not exist." % str(pkg_path))

    # And receate the directories and database based on the config file.
    logger.info("Recreating database and package dir.")
    _create_directories(server_path, config.PACKAGE_DIR, "/var/log/pynuget")
    _create_db(config.DB_BACKEND, config.DB_NAME, server_path)

    return False


def rebuild():
    """Rebuild the package database."""
    raise NotImplementedError
    import config
    # First let's get a list of all the packages in the database.
    # TODO: create the session.
    logger.debug("Getting database packages and versions.")
    db_data = db.search_packages(session, include_prerelease=True)
    db_data = _db_data_to_dict(db_data)

    # Then we'll get the list of all the packages in the package directory
    # Same data structure as the db data.
    pkg_path = Path(config.SERVER_PATH) / Path(config.PACKAGE_DIR)
    file_data = _get_packages_from_files(pkg_path)

    _add_packages_to_db(file_data)
    _remove_packages_from_db(file_data, db_data)

    return False


def push(file, source, key):
    """
    Push a package to a nuget server.

    Mimics the basic functionality of::

        nuget.exe push -ApiKey $key -Source $source $file

    This is needed because the standard NuGet.exe is not sending
    Content-Length in the header during HTTP PUT. I also can't figure out how
    to get Apache to ignore Content-Length (as it's a SHOULD not a MUST in
    the spec), and the Content-Length apache direcive isn't available until
    Apache 2.5...

    Parameters
    ----------
    file : str
        The path to the file to upload
    source : str
        The URL for the (py)NuGet server to push to.
    key : str
        The ApiKey value.
    """
    logger.debug("push('%s', '%s', '<redacted>')" % (file, source))

    if not Path(file).exists():
        logger.error("File '%s' does not exist. Aborting." % file)
        return

    header = {
        'X-Nuget-ApiKey': key,
        'User-Agent': 'PyNuGet',
    }

    if source[-1] == "/":
        source = source[:-1]
    source += '/api/v2/package/'

    files = {'package': open(file, 'rb')}
    resp = requests.put(source, headers=header, files=files)
    logger.debug("{} {}".format(resp, resp.text))

    # 201 = Item Created. Means we were successful.
    return resp.status_code == 201


def _create_dir(path):
    """
    Create (and chown) a given directory.

    Parameters
    ----------
    path : pathlib.Path

    Returns
    -------
    None
    """
    logger.debug("Creating '%s'" % path)
    try:
        # u=rwx,g=srwx,o=rx
        path.mkdir(parents=True, exist_ok=True)
    except PermissionError:
        logger.warn("Unable to make dir %s" % path)
        logger.warn(path.parent.stat())
        logger.warn("Parent Mode: %s" % (path.parent.stat().st_mode & 0o0777))
    else:
        try:
            shutil.chown(str(path), 'www-data', 'www-data')
        except PermissionError:
            logger.warn("Unable to change owner of %s" % path)


def _replace_prompt(path):
    """Return True if the user wants to replace the file."""
    while True:
        result = input("{} already exists. Replace? [yN] ".format(path))
        result = result.lower()
        if result in ('y', 'yes'):
            return True
        elif result in ('n', 'no', ''):
            return False
        else:
            print("Invalid answer. Please answer 'yes' or 'no'.")


def _now_str():
    """Return the current local time as a str for appending to a filename."""
    return dt.now().strftime("%y%m%d-%H%M%S")


def _copy_file_with_replace_prompt(src, dst, replace=None):
    """
    Copy a file, prompting to replace if it exists. Always save old file.

    Parameters
    ----------
    src : :class:`pathlib.Path`
        The source file to copy.
    dst : :class:`pathlib.Path`
        The location to copy to.
    replace : bool or None

    Returns
    -------
    modified : bool
        If True, the file was modified (created or replaced). If False, the
        existing file was not modified.
    """
    if replace is None:
        replace = _replace_prompt(dst)

    # Save the old file by renaming it with a timestamp.
    # TODO: I don't like this logic very much. There's got to be a better way.
    if dst.exists():
        logger.debug("Path {} already exists".format(dst))
        if replace:
            logger.debug("Replacing (and saving previous version)")
            dst.rename(Path(str(dst) + "." + _now_str()))
            shutil.copy(str(src.resolve()), str(dst))
            return True
        else:
            logger.debug("Not replacing")
            return False
    else:
        logger.debug("Copying new file to {}".format(dst))
        shutil.copy(str(src.resolve()), str(dst))
        return True


def _db_data_to_dict(db_data):
    """
    Convert the result of db.search_packages into a dict of
        {'pkg': ['vers1', 'vers2', ...]}
    """
    data = {}
    for row in db_data:
        try:
            data[row.package.name]
        except KeyError:
            data[row.package.name] = []
        data[row.package.name].append(row.version)

    logger.debug("Found %d database packages." % len(data))
    logger.debug("Found %d database versions." % sum(len(v) for v
                                                     in data.values()))
    return data


def _get_packages_from_files(pkg_path):
    """
    Get a list of packages from the package directory.

    Parameters
    ----------
    pkg_path : :class:`pathlib.Path` or str
        The path to the package directory.

    Returns
    -------
    data : dict
        Dict of {'pkg_name': ['vers1', 'vers2', ...], ...}
    """
    logger.debug("Getting list of packages in package dir.")
    if not isinstance(pkg_path, Path):
        pkg_path = Path(pkg_path)

    data = {}
    # XXX: There's got to be a better way!
    for root, dirs, _ in os.walk(str(pkg_path)):
        rel_path = Path(root).relative_to(pkg_path)
        pkg = str(rel_path.parent)
        if pkg != '.':
            try:
                data[pkg]
            except KeyError:
                data[pkg] = []
            data[pkg].append(rel_path.name)

    logger.debug("Found %d packages." % len(data))
    logger.debug("Found %d versions." % sum(len(v) for v in data.values()))

    return data


def _add_packages_to_db(file_data):
    logger.debug("Adding packages to database.")
    raise NotImplementedError
    for pkg, versions in file_data.items():
        # TODO
        # Check that the package exists in the database.
        if not package_in_db(pkg):
            db.insert_or_update_package(session, None, pkg, versions[0])

        for version in versions:
            if not version_in_db(pkg, version):
                db.insert_version(session, package_id=None, title=pkg,
                                  version=version)


def _remove_packages_from_db(file_data, db_data):
    logger.debug("Removing packages from database.")
    raise NotImplementedError
    for pkg, versions in db_data.items():
        if pkg not in file_data.keys():
            db.delete_version(pkg)
        else:
            for version in versions:
                if version not in file_data[pkg]:
                    db.delete_version(pkg, version)


def _check_permissions():
    """Raise PermissionError if we're not root/sudo."""
    if os.getuid() != 0:
        logger.warn("This script probably needs `sudo`. Trying anyway.")


def _set_permissions(paths):
    """
    Set permissions because apparently nothing else likes to..."

    Parameters
    ----------
    paths : iterable of str or iterable of pathlib.Path objects
    """
    logger.debug("Setting owner and permissions for %s" % paths)
    for path in paths:
        try:
            args = ['chown', '-R', 'www-data:www-data', str(path)]
            subprocess.run(args, check=True)
        except subprocess.CalledProcessError as err:
            logger.exception("Unlable to chown %s" % path)

        try:
            args = ['chmod', '-R', '2775', str(path)]
            subprocess.run(args, check=True)
        except subprocess.CalledProcessError as err:
            logger.exception("Unlable to chmod %s" % path)


def _create_directories(server_path, package_dir, log_dir):
    """Create the server directories if they don't exist."""
    logger.info("Creating directories (if they don't exist).")
    server_path = Path(server_path)
    package_dir = Path(package_dir)
    log_path = Path(log_dir)

    if not server_path.is_absolute():
        server_path = Path.cwd() / server_path
        logger.warn("'server_path' is not absolute, setting to %s" % server_path)

    if not package_dir.is_absolute():
        package_dir = server_path / package_dir
        logger.warn("'package_dir' is not absolue, setting to %s" % package_dir)

    if not log_path.is_absolute():
        log_path = Path.cwd() / log_dir
        logger.warn("'log_dir' is not absolute, setting to %s" % log_path)

    _create_dir(server_path)
    _create_dir(package_dir)
    _create_dir(log_path)


def _create_db(db_backend, db_name, server_path):
    """Create the database (file or schema) if it doesn't exist."""
    logger.info("Creating database.")
    logger.debug("db_backend={}, db_name={}".format(db_backend, db_name))

    url = None

    if db_backend == 'sqlite':
        db_name = Path(server_path) / Path(db_name)
        url = "sqlite:///{}".format(str(db_name))
        logger.debug(url)

        if db_name.exists():
            # assume that the sqlite file has the correct structure.
            pass
        else:
            # create the sqlite file and database.
            engine = create_engine(url, echo=False)
            db.Base.metadata.create_all(engine)
#            shutil.chown(str(db_name), 'www-data', 'www-data')
            os.chmod(str(db_name), 0o0664)
    elif db_backend in ('mysql', 'postgresql'):
        msg = "The backend '%s' is not yet implmented" % db_backend
        logger.error('Other database backends are not yet supported')
        raise NotImplementedError(msg)
    else:
        msg = "Invalid `db_backend` value: %s. Must be one of %s."
        msg = msg % (db_backend, ('sqlite', 'mysql', 'postgresql'))
        logger.error(msg)
        raise ValueError(msg)


def _copy_wsgi(server_path, replace_existing=None):
    """
    Copy the WSGI file to the server directory.

    Parameters
    ----------
    server_path : str
    replace_existing : bool or None
        If None, prompt the user to replace. If True, rename the old file
        and make a new one. If False, do nothing.

    Returns
    -------
    None
    """
    logger.info("Copying WSGI file.")
    original = Path(sys.prefix) / 'data' / 'wsgi.py'
    wsgi_path = Path(server_path) / 'wsgi.py'

    _copy_file_with_replace_prompt(original, wsgi_path, replace_existing)

    return wsgi_path


def _update_wsgi(wsgi_path):
    """
    Update the WSGI file with the correct venv directory.

    Parameters
    ----------
    wsgi_path : Path object
    """
    logger.debug("Updating WSGI file: %s" % wsgi_path)
    # Find our site-packages folder. Don't do anything if we're not in a venv.
    venv = os.getenv('VIRTUAL_ENV', None)
    if venv is None:
        logger.info("Not in a virtual env. Nothing to do.")
        return

    venv = Path(venv)
    site_pkgs = list(venv.glob('lib/python*/site-packages'))[0]
    logger.debug("Found site-packages directory: %s" % site_pkgs)

    args = ["sed",
            "-i", "-E",
            r's@^(site\.addsitedir).+$@\1\("{}"\)@g'.format(site_pkgs),
            str(wsgi_path),
            ]
    logger.debug("Command: %s" % " ".join(args))
    try:
        subprocess.run(args)
    except subprocess.CalledProcessError:
        msg = ("Unlable update the WSGI file with the virtual environment's"
               " site-packages directory! Please do so manually")
        logger.error(msg)


def _copy_apache_config(apache_config, replace_existing=None):
    """
    Copy the example apache config to the Apache sites.

    Parameters
    ----------
    apache_config : str
        Name of the apache config file. Must not be an absolute path.
    replace : bool or None
        If None, prompt the user to replace the config file. If the config
        file is overwritten (True or user responds True, the old file will
        still be saved). If False, the default config is not copied.

    Returns
    -------
    apache_conf : :class:`pathlib.Path`
        Full path to the apache config file.
    """
    logger.info("Copying example Apache Config.")
    apache_path = Path('/etc/apache2/sites-available/')
    apache_config = Path(apache_config)
    if apache_config.is_absolute():
        raise OSError("'apache_config' must not be an absolue path")
    apache_conf = apache_path / apache_config
    example_conf = Path(sys.prefix) / 'data' / 'apache-example.conf'

    _copy_file_with_replace_prompt(example_conf, apache_conf, replace_existing)

    return apache_conf


def _enable_apache_conf(conf):
    """
    Enable the apache site.

    Enabling a site simply consists of symlinking to the sites-available
    directory.

    Parameters
    ----------
    conf : :class:`pathlib.Path`
        The full path to the apache config file.

    Returns
    -------
    link : :class:`pathlib.Path`
        The path to the symlink.
    """
    logger.info("Enabling Apache site: {}".format(conf))

    link = conf.parent.parent / 'sites-enabled' / conf.name

    try:
        link.symlink_to(conf)
    except FileExistsError:
        logger.info("Site is already enabled.")

    return link


def _reload_apache():
    """Reload the apache configuration."""
    logger.info("Reloading Apache configuration")
    try:
        args = ['service', 'apache2', 'restart']
        subprocess.run(args, check=True)
    except subprocess.CalledProcessError as err:
        msg = ("Unlable to reload the apache configuration. Please attempt to"
               " reload it manually.")
        logger.error(err)
        logger.error(msg)


def _save_config(default_config_file, save_to=None, **kwargs):
    """Save the values to the configuration file."""
    logger.info("Saving configuration.")

    # Open the default config file.
    with open(str(default_config_file), 'r') as openf:
        raw = openf.read()

        for variable, new_value in kwargs.items():
            global_variable = variable.upper()

            pat = r'''^(?P<variable>{} = )['"](?P<value>.+)['"]$'''
            pat = re.compile(pat.format(global_variable), re.MULTILINE)

            old_value = pat.search(raw)
            if old_value is None:
                logger.debug("ignoring {}".format(global_variable))
                continue

            old_value = old_value.group('value')
            if old_value == new_value:
                logger.debug("Parameter '{}' unchanged.".format(variable))
                continue

            raw = pat.sub('\g<variable>"{}"'.format(new_value), raw)
            new_value = pat.search(raw).group('value')
            msg = "Replaced %s: '%s' with '%s'"
            logger.debug(msg % (variable, old_value, new_value))

    config_path = save_to
    if save_to is None:
        config_path = Path(kwargs['server_path']) / Path('config.py')

    with open(str(config_path), 'w') as openf:
        openf.write(raw)

    logger.debug("Configuration saved to `%s`" % config_path.as_posix())



# -*- coding: utf-8 -*-
"""
"""
import base64
import hashlib
import os
import re
from pathlib import Path
from zipfile import ZipFile

from flask import current_app
from lxml import etree as et

from pynuget import logger


class PyNuGetException(Exception):
    pass


class ApiException(PyNuGetException):
    pass


def require_auth(headers):
    """Ensure that the API key is valid."""
    key = headers.get('X-Nuget-Apikey', None)
    is_valid = key is not None and key in current_app.config['API_KEYS']
    if not is_valid:
        logger.error("Missing or Invalid API key")
    return is_valid


def get_package_path(package_id, version):
    """
    Get the file path for the specified pkg version.

    Parameters
    ----------
    package_id : str
        The name of the package
    version : str
        The package version

    Returns
    -------
    :class:`pathlib.Path`
        Relative path to the package file.
    """
    result = Path(package_id) / (version + '.nupkg')
    return result


def et_to_str(node):
    """Get the text value of an Element, returning None if not found."""
    try:
        return node.text
    except AttributeError:
        return None


def determine_dependencies(metadata_element, namespace):
    """"""
    # TODO: python-ify
    logger.debug("Parsing dependencies.")
    dependencies = []
    dep = metadata_element.find('nuspec:dependencies', namespace)
    if et.iselement(dep):
        logger.debug("Found dependencies.")
        dep_no_fw = dep.findall('nuspec:dependency', namespace)
        if dep_no_fw:
            logger.debug("Found dependencies not specific to any framework.")
            for dependency in dep_no_fw:
                d = {'framework': None,
                     'id': str(dependency.attrib['id']),
                     'version': str(dependency.attrib['version']),
                     }
                dependencies.append(d)
        dep_fw = dep.findall('nuspec:group', namespace)
        if dep_fw:
            logger.debug("Found dependencies specific to a framework")
            for group in dep_fw:
                group_elem = group.findall('nuspec:dependency', namespace)
                for dependency in group_elem:
                    d = {'framework': str(group.attrib['targetFramework']),
                         'id': str(dependency.attrib['id']),
                         'version': str(dependency.attrib['version']),
                         }
                    dependencies.append(d)
    else:
        logger.debug("No dependencies found.")

    return dependencies


def hash_and_encode_file(file):
    """
    Parameters
    ----------
    file : :class:`pathlib.Path` object

    Returns:
    --------
    hash_ : bytes
    filesize : int
    """
    return encode_file(file, hash_file(file, hashlib.sha512))


def hash_file(file, algorithm=hashlib.md5):
    """
    Parameters
    ----------
    file : :class:`pathlib.Path` object
    algorithm : Hash algorithm constructor
        One of the hash algorithms present in the `hashlib` module.

    Returns:
    --------
    hash_ : bytes

    Note that the returned `hash_` value is in binary. To make it a human
    readable string, use `binascii.hexlify(hash_).decode('utf-8')`.
    """
    file = str(file)
    logger.debug("Hashing file %s" % file)
    m = algorithm()
    with open(file, 'rb') as openf:
        m.update(openf.read())
    hash_ = m.hexdigest()

    logger.debug("%s hash: %s, %s" % (algorithm.__name__, hash_, file))

    return m.digest()


def encode_file(file, hash_):
    """
    Parameters
    ----------
    file : :class:`pathlib.Path` object
    hash_ : bytes
        The value returned by `hash_file()`.

    Returns:
    --------
    hash_ : bytes
    filesize : int
    """
    logger.debug("Encoding in Base64.")
    hash_ = base64.b64encode(hash_)

    # Get the filesize of the uploaded file. Used later.
    filesize = os.path.getsize(file)
    logger.debug("File size: %d bytes" % filesize)

    return hash_.decode('utf-8'), filesize


def save_file(file, pkg_name, version):
    """
    Parameters
    ----------
    file : :class:`werkzeug.datastructures.FileStorage` object
        The file as retrieved by Flask.
    pkg_name : str
    version : str

    Returns:
    --------
    local_path : :class:`pathlib.Path`
        The path to the saved file.
    """
    # Save the package file to the local package dir. Thus far it's
    # just been floating around in magic Flask land.
    server_path = Path(current_app.config['SERVER_PATH'])
    package_dir = Path(current_app.config['PACKAGE_DIR'])
    local_path = server_path / package_dir
    local_path = local_path / pkg_name / (version + ".nupkg")

    # Check if the package's directory already exists. Create if needed.
    create_parent_dirs(local_path)

    logger.debug("Saving uploaded file to filesystem.")
    try:
        file.save(str(local_path))
    except Exception as err:       # TODO: specify exceptions
        logger.error("Unknown exception: %s" % err)
        raise err
    else:
        logger.info("Succesfully saved package to '%s'" % str(local_path))

    return local_path


def create_parent_dirs(path):
    """
    Create the parent directories if it doesn't already exist.

    Parameters
    ----------
    path : :class:`pathlib.Path`
    """
    os.makedirs(str(path.parent),
                mode=0o0755,
                exist_ok=True,      # do not throw an error path exists.
                )


def extract_nuspec(file):
    """
    Parameters
    ----------
    file : :class:`pathlib.Path` object or str
        The file as retrieved by Flask.
    """
    pkg = ZipFile(str(file), 'r')
    logger.debug("Parsing uploaded file.")
    nuspec_file = None
    pattern = re.compile(r'^.*\.nuspec$', re.IGNORECASE)
    nuspec_file = list(filter(pattern.search, pkg.namelist()))
    if len(nuspec_file) > 1:
        logger.error("Multiple NuSpec files found within the package.")
        raise ApiException("api_error: multiple nuspec files found")
    elif len(nuspec_file) == 0:
        logger.error("No NuSpec file found in the package.")
        raise ApiException("api_error: nuspec file not found")      # TODO
    nuspec_file = nuspec_file[0]

    with pkg.open(nuspec_file, 'r') as openf:
        nuspec_string = openf.read()
        logger.debug("NuSpec string:")
        logger.debug(nuspec_string)

    logger.debug("Parsing NuSpec file XML")
    nuspec = et.fromstring(nuspec_string)
    if not et.iselement(nuspec):
        msg = "`nuspec` expected to be type `xml...Element`. Got {}"
        raise TypeError(msg.format(type(nuspec)))

    return nuspec


def extract_namespace(nuspec):
    """
    Extract the namespce from the NuSpec file.

    Parameters
    ----------
    nuspec : :class:`lxml.etree.Element` object
        The parsed nuspec data.

    Returns
    -------
    namespace : str
    """
    namespace = nuspec.xpath('namespace-uri(.)')
    logger.debug("Found namespace: %s" % namespace)
    return namespace


def parse_nuspec(nuspec, ns=None):
    """
    Parameters
    ----------
    nuspec : :class:`lxml.etree.Element` object
        The parsed nuspec data.
    ns : string
        The namespace to search.

    Returns
    -------
    metadata : :class:`lxml.etree.Element`
    pkg_name : str
    version : str
    """
    metadata = nuspec.find('nuspec:metadata', ns)
    if metadata is None:
        msg = 'Unable to find the metadata tag!'
        logger.error(msg)
        raise ApiException(msg)

    # TODO: I think I need different error handling around `.text`.
    pkg_name = metadata.find('nuspec:id', ns)
    version = metadata.find('nuspec:version', ns)
    if pkg_name is None or version is None:
        logger.error("ID or version missing from NuSpec file.")
        raise ApiException("api_error: ID or version missing")        # TODO

    return metadata, pkg_name.text, version.text



# -*- coding: utf-8 -*-
"""
"""

import datetime as dt
import json
import re

from lxml import etree as et

from pynuget import logger

BASE = b"""<?xml version="1.0" encoding="utf-8" ?>
<feed
  xml:base="https://www.nuget.org/api/v2/"
  xmlns="http://www.w3.org/2005/Atom"
  xmlns:d="http://schemas.microsoft.com/ado/2007/08/dataservices"
  xmlns:m="http://schemas.microsoft.com/ado/2007/08/dataservices/metadata"
>
</feed>
"""
NS_D = "{http://schemas.microsoft.com/ado/2007/08/dataservices}"
NS_M = "{http://schemas.microsoft.com/ado/2007/08/dataservices/metadata}"
NS_DEFAULT = "{http://www.w3.org/2005/Atom}"
NSMAP = {
    None: NS_DEFAULT.strip("{}"),
    'd': NS_D.strip("{}"),
    'm': NS_M.strip("{}"),
}

ADO_BASE_URL = "http://schemas.microsoft.com/ado/2007/08/dataservices"
ADO_SCHEMA_URL = ADO_BASE_URL + "/scheme"
ADO_METADATA_URL = ADO_BASE_URL + "/metadata"


class FeedWriter(object):

    def __init__(self, feed_name, base_url='http://localhost:5000/'):
        logger.debug("Initializing FeedWriter")
        self.feed_name = feed_name
        self.base_url = base_url

    def write(self, results):
        self.begin_feed()
        try:
            for result in results:
                self.add_entry(result)
        except TypeError:
            pass
        return et.tostring(self.feed)

    def write_to_output(self, results):
        """
        results : list of dicts, I think.
        """
        logger.debug("FeedWriter.write_to_output(%d)" % len(results))
        # TODO: header line
        return self.write(results)

    def begin_feed(self):
        logger.debug("FeedWriter.begin_feed()")
        self.feed = et.fromstring(BASE)
        node = et.Element('id')
        node.text = self.base_url + str(self.feed_name)
        self.feed.append(node)
        self.add_with_attributes(
            self.feed,
            'title',
            self.feed_name,
            {'type': 'text'},
        )
        node = et.Element('updated')
        node.text = self.format_date(dt.datetime.utcnow())
        self.feed.append(node)
        self.add_with_attributes(
            self.feed,
            'link',
            None,
            {'rel': 'self', 'title': self.feed_name, 'href': self.feed_name},
        )

    def add_entry(self, row):
        """
        Parameters
        ----------
        row :
            SQLAlchemy result set object
        """
        logger.debug("FeedWriter.add_entry(%s)" % row)
        entry_id = 'Packages(Id="{}",Version="{}")'.format(row.package_id,
                                                           row.version)
        entry = et.Element('entry')
        self.feed.append(entry)
        node = et.Element('id')
        node.text = self.base_url + entry_id
        entry.append(node)
        self.add_with_attributes(
            entry,
            'category',
            None,
            {'term': 'NuGetGallery.V2FeedPackage', 'scheme': ADO_SCHEMA_URL},
        )
        self.add_with_attributes(
            entry,
            'link',
            None,
            {'rel': 'edit', 'title': 'V2FeedPackage', 'href': entry_id},
        )

        # Yes, this "title" is actually the package ID. Actual title is in
        # the metadata.
        self.add_with_attributes(entry, 'title', str(row.package_id),
                                 {'type': 'text'})
        self.add_with_attributes(entry, 'summary', None, {'type': 'text'})
        node = et.Element('updated')
        node.text = self.format_date(row.created)
        entry.append(node)

        authors_node = et.Element('author')
        entry.append(authors_node)
        node = et.Element('name')
        node.text = row.authors
        authors_node.append(node)

        self.add_with_attributes(
            entry,
            'link',
            None,
            {'rel': 'edit-media',
             'title': 'V2FeedPackage',
             'href': entry_id + '/$value',
             },
        )
        url = "{}download/{}/{}".format(self.base_url, row.package_id,
                                        row.version)
        self.add_with_attributes(
            entry,
            'content',
            None,
            {'type': 'application/zip', 'src': url},
        )
        self.add_entry_meta(entry, row)

    def add_entry_meta(self, entry, row):
        """
        Parameters
        ----------
        row :
            SQLAlchemy result set object
        """
        logger.debug("FeedWriter.add_entry_meta(%s, %s)" % (entry, row))
        properties = et.Element(NS_M + 'properties', nsmap=NSMAP)
        entry.append(properties)

        gallery_details_url = "{}details/{}/{}".format(self.base_url,
                                                       row.package_id,
                                                       row.version)

        meta = {
            'Id': row.title,
            'Version': row.version,
            'NormalizedVersion': row.version,
            'Copyright': row.copyright_,
            'Created': self.render_meta_date(row.created),
            'Dependencies': self.render_dependencies_xml(row.dependencies),
            'Description': row.description,  # TODO: htmlspecialchars
            'DownloadCount': {'value': str(row.package.download_count), 'type': 'Edm.Int32'},
            'GalleryDetailsUrl': gallery_details_url,
            'IconUrl': row.icon_url,  #TODO: htmlspecialchars
            'IsLatestVersion': self.render_meta_boolean(row.package.latest_version == row.version),
            'IsAbsoluteLatestVersion': self.render_meta_boolean(row.package.latest_version == row.version),
            'IsPrerelease': self.render_meta_boolean(row.is_prerelease),
            'Language': None,
            'Published': self.render_meta_date(row.created),
            'PackageHash': row.package_hash,
            'PackageHashAlgorithm': row.package_hash_algorithm,
            'PackageSize': {'value': str(row.package_size), 'type': 'Edm.Int64'},
            'ProjectUrl': row.project_url,
            'ReportAbuseUrl': '',
            'ReleaseNotes': row.release_notes,  # TODO: htmlspecialchars
            'RequireLicenseAcceptance': self.render_meta_boolean(row.require_license_acceptance),
            'Summary': None,
            'Tags': row.tags,
            'Title': row.title,
            'VersionDownloadCount': {'value': str(row.version_download_count), 'type': 'Edm.Int32'},
            'MinClientVersion': '',
            'LastEdited': {'value': None, 'type': 'Edm.DateTime'},
            'LicenseUrl': row.license_url,
            'LicenseNames': '',
            'LicenseReportUrl': '',
        }

        for name, data in sorted(meta.items()):
            if isinstance(data, et._Element):
                properties.append(data)
                continue

            if isinstance(data, dict):
                value = data['value']
                type_ = data['type']
            else:
                value = data
                type_ = None

            self.add_meta(properties, name, value, type_)

    def render_meta_date(self, date):
        return {'value': self.format_date(date) + "Z", 'type': 'Edm.DateTime'}

    def render_meta_boolean(self, value):
        return {'value': str(value), 'type': 'Edm.Boolean'}

    def format_date(self, value):
        logger.debug("{}: {}".format(value, type(value)))
        if isinstance(value, dt.datetime):
            #  return value.isoformat(timespec='seconds')      # Py3.6+
            return value.isoformat()
        return value

    def render_dependencies(self, raw):
        logger.debug("FeedWriter.render_dependencies(%s)" % raw)
        if not raw:
            return ''

        try:
            data = json.loads(raw)
        except json.decoder.JSONDecodeError:
            return ''

        output = []

        for dependency in data:
            formatted_dependency = "{}:{}:".format(dependency['id'],
                                                   dependency['version'])
            if 'framework' in dependency.keys():
                formatted_dependency += self.format_target_framework(
                    dependency['framework'],
                )
            output.append(formatted_dependency)

        return "|".join(output)

    def render_dependencies_xml(self, raw):
        """
        Convert a raw json in the format:
            [
                {"version": "4.0.0", "framework": null, "id": "NLog"},
                {"version": "4.0.0", "framework": "A", "id": "pkg2"},
                {"version": "4.0.0", "framework": "A", "id": "pkg1"},
                {"version": "4.0.0", "framework": "B", "id": "pkg3"},
            ]

        to the following XML:
            <group>
              <dependency id="NLog" version="4.0.0" />
            </group>
            <group targetFramework="A">
              <dependency id="pkg2" version="4.0.0" />
              <dependency id="pkg1" version="4.0.0" />
            </group>
            <group targetFramework="B">
              <dependency id="pkg3" version="4.0.0" />
            </group>
        """
        logger.debug("FeedWriter.render_dependencies(%s)" % raw)
        if not raw:
            return ''

        try:
            data = json.loads(raw)
        except json.decoder.JSONDecodeError:
            return ''

        data = group_dependencies(data)

        # then render the xml.
        root_node = et.Element(NS_D + 'Dependencies', nsmap=NSMAP)
        for group, deps in data.items():
            node = et.Element(NS_D + 'Group', nsmap=NSMAP)
            if group is not None:
                node.set('targetFramework', group)
            for dep in deps:
                dep_node = et.Element(NS_D + 'Dependency', nsmap=NSMAP)
                dep_node.set('id', dep['id'])
                dep_node.set('version', dep['version'])
                node.append(dep_node)
            root_node.append(node)

        return root_node

    def format_target_framework(self, framework):
        """Format a raw target framework from a NuSpec into the format
        used in the packages feed.

        Eg:
        DNX4.5.1 -> dnx451
        DNXCore5.0 -> dnxcore50
        """
        return re.sub('[^A-Z0-9]', '', framework, flags=re.IGNORECASE).lower()

    def add_with_attributes(self, entry, name, value, attributes):
        """

        Parameters
        ----------
        entry :
            Looks to be a SimpleXMLElement node object
        attributes :
            Dict, I think.
        """
        child = et.Element(name)
        child.text = str(value)
        entry.append(child)
        for attr_name, attr_value in sorted(attributes.items()):
            child.set(attr_name, str(attr_value))

    def add_meta(self, entry, name, value, type_=None):

        child = et.Element(NS_D + name, nsmap=NSMAP)
        child.text = value
        entry.append(child)

        if type_:
            child.set('type', type_)

        if value is None:
            child.set('null', 'true')


def group_dependencies(data):
    """
    Groups each dependency into a dict of items with the same framework.

    Parameters
    ----------
    data :
        Parsed JSON

    Returns
    -------
    groups : dict
    """
    logger.debug("Starting group_dependencies(...)")
    groups = {}
    for item in data:
        framework = item.get('framework', None)

        # special case: 'null' framework
        if framework == 'null':
            framework = None

        try:
            groups[framework].append(item)
        except KeyError:
            groups[framework] = [item]

    grp_names = groups.keys()
    logger.debug("Found {} dependency groups: {}".format(len(grp_names),
                                                         grp_names))
    return groups



# -*- coding: utf-8 -*-
"""
"""
import os
from pathlib import Path

from flask import Flask
from flask import g

from pynuget import logger
from pynuget._logging import setup_logging
from pynuget.routes import pages

ENV_VAR = "PYNUGET_CONFIG_TYPE"


def create_app():
    """
    Application Factory.
    """

    instance_path = Path("/var/www/pynuget")
    config_file = instance_path / "config.py"

    testing = os.getenv(ENV_VAR, None) == 'TESTING'

    app = Flask(__name__)
    app.config.from_object('pynuget.default_config')

    # Override some default vars if we're running using the flask dev server.
    if os.getenv(ENV_VAR, None) == "LOCAL_DEV":
        app.config['LOG_PATH'] = os.getenv('PYNUGET_LOG_PATH')
        app.config['SERVER_PATH'] = os.getenv('PYNUGET_SERVER_PATH')

    if config_file.exists():
        logger.debug("Found config file: %s. Loading..." % str(config_file))
        app.config.from_pyfile(str(config_file))

    # Update logging to also log to a file.
    # This *should* modify the logger that was created in __init__.py...
    if not testing:
        setup_logging(to_console=False, to_file=True,
                      log_path=app.config['LOG_PATH'])

    # Register blueprints
    app.register_blueprint(pages)

    @app.teardown_appcontext
    def teardown_db_session(exception):
        session = getattr(g, 'session', None)
        if session is not None:
            session.close()

    return app



# -*- coding: utf-8 -*-
"""
"""
from setuptools import setup, find_packages
from setuptools.command.install import install
from setuptools.command.develop import develop

from glob import glob
from os.path import basename
from os.path import splitext

from pathlib import Path
import re


# #########################################################################
# These classes will only work if the user runs:
#    python setup.py develop
#    python setup.py install
class DevelopCommand(develop):
    """
    Modify the `develop` installation workflow.
    """
    def run(self):
        # Pre-install: nothing
        develop.run(self)
        # Post-install: perhaps init the dirs and database?


class InstallCommand(install):
    """
    Modify the starndard installation workflow.
    """
    def run(self):
        # Pre-install: nothing
        install.run(self)
        # Post-install: perhaps init the dirs and database?

# #########################################################################


def find_version(*file_paths):
    """
    Find the package version for the given path.

    Taken from https://github.com/pypa/pip/blob/master/setup.py and modified
    to suit my needs/tastes.
    """
    version_file = Path(__file__).parent.joinpath(*file_paths)
    with open(str(version_file), 'r') as openf:
        data = openf.read()
    version_match = re.search(
        r"^__version__ = ['\"]([^'\"]*)['\"]",
        data,
        re.M,
    )
    if version_match:
        return version_match.group(1)

    raise RuntimeError("Unable to find version string.")


classifiers = [
    "Development Status :: 4 - Beta",
    "Intended Audience :: Developers",
    "Intended Audience :: Information Technology",
    "Intended Audience :: System Administrators",
    "License :: OSI Approved :: MIT License",
    "Operating System :: POSIX :: Linux",
    "Programming Language :: Python :: 3.5",
    "Programming Language :: Python :: 3.6",
    "Programming Language :: Python :: 3.7",
    "Programming Language :: Python :: 3 :: Only",
    "Programming Language :: Python :: Implementation :: CPython",
    "Topic :: Internet :: WWW/HTTP :: WSGI :: Application",
    "Topic :: Software Development :: Build Tools",
]

requires = [
    "sqlalchemy>=1.2.7",
    "flask>=0.12.2",
    "lxml>=4.2.1",
    "requests>=2.19.1",
]

entry_points = {
    'console_scripts': [
        'pynuget = pynuget.cli:main',
    ],
}

with open("README.rst", 'r') as openf:
    long_description = openf.read()

setup(
    # General
    name="pynuget",
    version=find_version('src', 'pynuget', '__init__.py'),

    # Descriptions
    description="A Python-based NuGet Server",
    long_description=long_description,
    long_description_content_type='text/x-rst',

    # Owner
    url='https://github.com/dougthor42/pynuget/',
    author='Douglas Thor',
    author_email='doug.thor@gmail.com',

    # Files
    package_dir={'': 'src'},
    packages=find_packages(where='src'),
    include_package_data=True,

    # these get copied to sys.prefix/data, which will be the venv's folder.
    data_files=[
        ('data', ['wsgi.py']),
        ('data', ['apache-example.conf']),
    ],

    py_modules=[splitext(basename(path))[0] for path in glob('src/*.py')],

    classifiers=classifiers,
    keywords='nuget packaging c# csharp',

    # Versions and Requirements
    python_requires=">=3.5",
    install_requires=requires,
    #tests_requires=

    entry_points=entry_points,

    # https://stackoverflow.com/a/36902139/1354930
    cmdclass={
        'develop': DevelopCommand,
        'install': InstallCommand,
    },
)
